@@ -1,4 +1,5 @@
 import threading
+import os
 try:
     from collections import OrderedDict
 except ImportError:
@@ -7,6 +8,7 @@ except ImportError:
 from .pgcompleter import PGCompleter
 from .pgexecute import PGExecute
 
+
 class CompletionRefresher(object):
 
     refreshers = OrderedDict()
@@ -15,7 +17,8 @@ class CompletionRefresher(object):
         self._completer_thread = None
         self._restart_refresh = threading.Event()
 
-    def refresh(self, executor, special, callbacks):
+    def refresh(self, executor, special, callbacks, history=None,
+      settings=None):
         """
         Creates a PGCompleter object and populates it with the relevant
         completion suggestions in a background thread.
@@ -23,6 +26,7 @@ class CompletionRefresher(object):
         executor - PGExecute object, used to extract the credentials to connect
                    to the database.
         special - PGSpecial object used for creating a new completion object.
+        settings - dict of settings for completer object
         callbacks - A function or a list of functions to call after the thread
                     has completed the refresh. The newly created completion
                     object will be passed in as an argument to each callback.
@@ -31,9 +35,10 @@ class CompletionRefresher(object):
             self._restart_refresh.set()
             return [(None, None, None, 'Auto-completion refresh restarted.')]
         else:
-            self._completer_thread = threading.Thread(target=self._bg_refresh,
-                                                      args=(executor, special, callbacks),
-                                                      name='completion_refresh')
+            self._completer_thread = threading.Thread(
+                target=self._bg_refresh,
+                args=(executor, special, callbacks, history, settings),
+                name='completion_refresh')
             self._completer_thread.setDaemon(True)
             self._completer_thread.start()
             return [(None, None, None,
@@ -42,8 +47,10 @@ class CompletionRefresher(object):
     def is_refreshing(self):
         return self._completer_thread and self._completer_thread.is_alive()
 
-    def _bg_refresh(self, pgexecute, special, callbacks):
-        completer = PGCompleter(smart_completion=True, pgspecial=special)
+    def _bg_refresh(self, pgexecute, special, callbacks, history=None,
+      settings=None):
+        completer = PGCompleter(smart_completion=True, pgspecial=special,
+            settings=settings)
 
         # Create a new pgexecute method to popoulate the completions.
         e = pgexecute
@@ -68,9 +75,16 @@ class CompletionRefresher(object):
             # break statement.
             continue
 
+        # Load history into pgcompleter so it can learn user preferences
+        n_recent = 100
+        if history:
+            for recent in history[-n_recent:]:
+                completer.extend_query_history(recent, is_init=True)
+
         for callback in callbacks:
             callback(completer)
 
+
 def refresher(name, refreshers=CompletionRefresher.refreshers):
     """Decorator to populate the dictionary of refreshers with the current
     function.
@@ -89,6 +103,7 @@ def refresh_schemata(completer, executor):
 def refresh_tables(completer, executor):
     completer.extend_relations(executor.tables(), kind='tables')
     completer.extend_columns(executor.table_columns(), kind='tables')
+    completer.extend_foreignkeys(executor.foreignkeys())
 
 @refresher('views')
 def refresh_views(completer, executor):
@@ -106,3 +121,17 @@ def refresh_types(completer, executor):
 @refresher('databases')
 def refresh_databases(completer, executor):
     completer.extend_database_names(executor.databases())
+
+@refresher('casing')
+def refresh_casing(completer, executor):
+    casing_file = completer.casing_file
+    if not casing_file:
+        return
+    generate_casing_file = completer.generate_casing_file
+    if generate_casing_file and not os.path.isfile(casing_file):
+        casing_prefs = '\n'.join(executor.casing())
+        with open(casing_file, 'w') as f:
+            f.write(casing_prefs)
+    if os.path.isfile(casing_file):
+        with open(casing_file, 'r') as f:
+            completer.extend_casing([line.strip() for line in f])
