@@ -9,20 +9,24 @@ import traceback
 import logging
 import threading
 import shutil
+import functools
+import humanize
 from time import time
 from codecs import open
 
+
 import click
 try:
     import setproctitle
 except ImportError:
     setproctitle = None
-import sqlparse
 from prompt_toolkit import CommandLineInterface, Application, AbortAction
-from prompt_toolkit.enums import DEFAULT_BUFFER
-from prompt_toolkit.shortcuts import create_default_layout, create_eventloop
+from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
+from prompt_toolkit.shortcuts import create_prompt_layout, create_eventloop
+from prompt_toolkit.buffer import AcceptAction
 from prompt_toolkit.document import Document
 from prompt_toolkit.filters import Always, HasFocus, IsDone
+from prompt_toolkit.layout.lexers import PygmentsLexer
 from prompt_toolkit.layout.processors import (ConditionalProcessor,
                                         HighlightMatchingBracketProcessor)
 from prompt_toolkit.history import FileHistory
@@ -39,7 +43,8 @@ from .pgstyle import style_factory
 from .pgexecute import PGExecute
 from .pgbuffer import PGBuffer
 from .completion_refresher import CompletionRefresher
-from .config import write_default_config, load_config, config_location
+from .config import (get_casing_file,
+    load_config, config_location, ensure_dir_exists, get_config)
 from .key_bindings import pgcli_bindings
 from .encodingutils import utf8tounicode
 from .__init__ import __version__
@@ -47,9 +52,9 @@ from .__init__ import __version__
 click.disable_unicode_literals_warning = True
 
 try:
-    from urlparse import urlparse
+    from urlparse import urlparse, unquote
 except ImportError:
-    from urllib.parse import urlparse
+    from urllib.parse import urlparse, unquote
 
 from getpass import getuser
 from psycopg2 import OperationalError
@@ -71,63 +76,98 @@ MetaQuery = namedtuple(
 MetaQuery.__new__.__defaults__ = ('', False, 0, False, False, False, False)
 
 
+# no-op logging handler
+class NullHandler(logging.Handler):
+    def emit(self, record):
+        pass
+
+
 class PGCli(object):
 
+    def set_default_pager(self, config):
+        configured_pager = config['main'].get('pager')
+        os_environ_pager = os.environ.get('PAGER')
+
+        if configured_pager:
+            self.logger.info('Default pager found in config file: ' + '\'' + configured_pager + '\'')
+            os.environ['PAGER'] = configured_pager
+        elif os_environ_pager:
+            self.logger.info('Default pager found in PAGER environment variable: ' + '\'' + os_environ_pager + '\'')
+            os.environ['PAGER'] = os_environ_pager
+        else:
+            self.logger.info('No default pager found in environment. Using os default pager')
+        # Always set default set of less recommended options, they are ignored if pager is
+        # different than less or is already parameterized with their own arguments
+        os.environ['LESS'] = '-SRXF'
+
     def __init__(self, force_passwd_prompt=False, never_passwd_prompt=False,
-                 pgexecute=None, pgclirc_file=None):
+                 pgexecute=None, pgclirc_file=None, row_limit=None):
 
         self.force_passwd_prompt = force_passwd_prompt
         self.never_passwd_prompt = never_passwd_prompt
         self.pgexecute = pgexecute
 
-        from pgcli import __file__ as package_root
-        package_root = os.path.dirname(package_root)
+        # Load config.
+        c = self.config = get_config(pgclirc_file)
 
-        default_config = os.path.join(package_root, 'pgclirc')
-        write_default_config(default_config, pgclirc_file)
+        self.logger = logging.getLogger(__name__)
+        self.initialize_logging()
 
+        self.set_default_pager(c)
+        self.output_file = None
         self.pgspecial = PGSpecial()
 
-        # Load config.
-        c = self.config = load_config(pgclirc_file, default_config)
         self.multi_line = c['main'].as_bool('multi_line')
         self.vi_mode = c['main'].as_bool('vi')
         self.pgspecial.timing_enabled = c['main'].as_bool('timing')
+        if row_limit is not None:
+            self.row_limit = row_limit
+        else:
+            self.row_limit = c['main'].as_int('row_limit')
+
         self.table_format = c['main']['table_format']
         self.syntax_style = c['main']['syntax_style']
         self.cli_style = c['colors']
         self.wider_completion_menu = c['main'].as_bool('wider_completion_menu')
+        self.less_chatty = c['main'].as_bool('less_chatty')
 
         self.on_error = c['main']['on_error'].upper()
 
         self.completion_refresher = CompletionRefresher()
 
-        self.logger = logging.getLogger(__name__)
-        self.initialize_logging()
-
         self.query_history = []
 
         # Initialize completer
         smart_completion = c['main'].as_bool('smart_completion')
-        completer = PGCompleter(smart_completion, pgspecial=self.pgspecial)
+        settings = {'casing_file': get_casing_file(c),
+            'generate_casing_file': c['main'].as_bool('generate_casing_file'),
+            'asterisk_column_order': c['main']['asterisk_column_order']}
+        completer = PGCompleter(smart_completion, pgspecial=self.pgspecial,
+            settings=settings)
         self.completer = completer
         self._completer_lock = threading.Lock()
         self.register_special_commands()
 
+        self.eventloop = create_eventloop()
         self.cli = None
 
     def register_special_commands(self):
 
-        self.pgspecial.register(self.change_db, '\\c',
-                              '\\c[onnect] database_name',
-                              'Change to a new database.',
-                              aliases=('use', '\\connect', 'USE'))
-        self.pgspecial.register(self.refresh_completions, '\\#', '\\#',
-                              'Refresh auto-completions.', arg_type=NO_QUERY)
-        self.pgspecial.register(self.refresh_completions, '\\refresh', '\\refresh',
-                              'Refresh auto-completions.', arg_type=NO_QUERY)
+        self.pgspecial.register(
+            self.change_db, '\\c', '\\c[onnect] database_name',
+            'Change to a new database.', aliases=('use', '\\connect', 'USE'))
+
+        refresh_callback = lambda: self.refresh_completions(
+            persist_priorities='all')
+
+        self.pgspecial.register(refresh_callback, '\\#', '\\#',
+                                'Refresh auto-completions.', arg_type=NO_QUERY)
+        self.pgspecial.register(refresh_callback, '\\refresh', '\\refresh',
+                                'Refresh auto-completions.', arg_type=NO_QUERY)
         self.pgspecial.register(self.execute_from_file, '\\i', '\\i filename',
-                              'Execute commands from file.')
+                                'Execute commands from file.')
+        self.pgspecial.register(self.write_to_file, '\\o', '\\o [filename]',
+                                'Send all query results to file.')
 
     def change_db(self, pattern, **_):
         if pattern:
@@ -142,30 +182,59 @@ class PGCli(object):
     def execute_from_file(self, pattern, **_):
         if not pattern:
             message = '\\i: missing required argument'
-            return [(None, None, None, message)]
+            return [(None, None, None, message, '', False)]
         try:
             with open(os.path.expanduser(pattern), encoding='utf-8') as f:
                 query = f.read()
         except IOError as e:
-            return [(None, None, None, str(e))]
+            return [(None, None, None, str(e), '', False)]
 
-        return self.pgexecute.run(query, self.pgspecial, on_error=self.on_error)
+        on_error_resume = (self.on_error == 'RESUME')
+        return self.pgexecute.run(
+            query, self.pgspecial, on_error_resume=on_error_resume
+        )
+
+    def write_to_file(self, pattern, **_):
+        if not pattern:
+            self.output_file = None
+            message = 'File output disabled'
+            return [(None, None, None, message, '', True)]
+        filename = os.path.abspath(os.path.expanduser(pattern))
+        if not os.path.isfile(filename):
+            try:
+                open(filename, 'w').close()
+            except IOError as e:
+                self.output_file = None
+                message = str(e) + '\nFile output disabled'
+                return [(None, None, None, message, '', False)]
+        self.output_file = filename
+        message = 'Writing to file "%s"' % self.output_file
+        return [(None, None, None, message, '', True)]
 
     def initialize_logging(self):
 
         log_file = self.config['main']['log_file']
         if log_file == 'default':
             log_file = config_location() + 'log'
+        ensure_dir_exists(log_file)
         log_level = self.config['main']['log_level']
 
+        # Disable logging if value is NONE by switching to a no-op handler.
+        # Set log level to a high value so it doesn't even waste cycles getting called.
+        if log_level.upper() == 'NONE':
+            handler = NullHandler()
+        else:
+            handler = logging.FileHandler(os.path.expanduser(log_file))
+
         level_map = {'CRITICAL': logging.CRITICAL,
                      'ERROR': logging.ERROR,
                      'WARNING': logging.WARNING,
                      'INFO': logging.INFO,
-                     'DEBUG': logging.DEBUG
+                     'DEBUG': logging.DEBUG,
+                     'NONE': logging.CRITICAL
                      }
 
-        handler = logging.FileHandler(os.path.expanduser(log_file))
+        log_level = level_map[log_level.upper()]
 
         formatter = logging.Formatter(
             '%(asctime)s (%(process)d/%(threadName)s) '
@@ -175,19 +244,25 @@ class PGCli(object):
 
         root_logger = logging.getLogger('pgcli')
         root_logger.addHandler(handler)
-        root_logger.setLevel(level_map[log_level.upper()])
+        root_logger.setLevel(log_level)
 
         root_logger.debug('Initializing pgcli logging.')
         root_logger.debug('Log file %r.', log_file)
 
+        pgspecial_logger = logging.getLogger('pgspecial')
+        pgspecial_logger.addHandler(handler)
+        pgspecial_logger.setLevel(log_level)
+
     def connect_dsn(self, dsn):
         self.connect(dsn=dsn)
 
     def connect_uri(self, uri):
         uri = urlparse(uri)
         database = uri.path[1:]  # ignore the leading fwd slash
-        self.connect(database, uri.hostname, uri.username,
-                     uri.port, uri.password)
+        arguments = [database, uri.hostname, uri.username,
+                     uri.port, uri.password]
+        # unquote each URI part (they may be percent encoded)
+        self.connect(*list(map(lambda p: unquote(str(p)) if p else p, arguments)))
 
     def connect(self, database='', host='', user='', port='', passwd='',
                 dsn=''):
@@ -269,20 +344,25 @@ class PGCli(object):
 
     def run_cli(self):
         logger = self.logger
-        original_less_opts = self.adjust_less_opts()
 
-        self.refresh_completions()
+        history_file = self.config['main']['history_file']
+        if history_file == 'default':
+            history_file = config_location() + 'history'
+        history = FileHistory(os.path.expanduser(history_file))
+        self.refresh_completions(history=history,
+                                 persist_priorities='none')
 
-        self.cli = self._build_cli()
+        self.cli = self._build_cli(history)
 
-        print('Version:', __version__)
-        print('Chat: https://gitter.im/dbcli/pgcli')
-        print('Mail: https://groups.google.com/forum/#!forum/pgcli')
-        print('Home: http://pgcli.com')
+        if not self.less_chatty:
+            print('Version:', __version__)
+            print('Chat: https://gitter.im/dbcli/pgcli')
+            print('Mail: https://groups.google.com/forum/#!forum/pgcli')
+            print('Home: http://pgcli.com')
 
         try:
             while True:
-                document = self.cli.run()
+                document = self.cli.run(True)
 
                 # The reason we check here instead of inside the pgexecute is
                 # because we want to raise the Exit exception which will be
@@ -325,19 +405,35 @@ class PGCli(object):
                     click.secho(str(e), err=True, fg='red')
                 else:
                     try:
-                        click.echo_via_pager('\n'.join(output))
+                        if self.output_file and not document.text.startswith(('\\o ', '\\? ')):
+                            try:
+                                with open(self.output_file, 'a', encoding='utf-8') as f:
+                                    click.echo(document.text, file=f)
+                                    click.echo('\n'.join(output), file=f)
+                                    click.echo('', file=f) # extra newline
+                            except IOError as e:
+                                click.secho(str(e), err=True, fg='red')
+                        else:
+                            click.echo_via_pager('\n'.join(output))
                     except KeyboardInterrupt:
                         pass
 
                     if self.pgspecial.timing_enabled:
-                        print('Time: %0.03fs' % query.total_time)
+                        # Only add humanized time display if > 1 second
+                        if query.total_time > 1:
+                            print('Time: %0.03fs (%s)' % (query.total_time,
+                                  humanize.time.naturaldelta(query.total_time)))
+                        else:
+                            print('Time: %0.03fs' % query.total_time)
 
                     # Check if we need to update completions, in order of most
                     # to least drastic changes
                     if query.db_changed:
-                        self.refresh_completions(reset=True)
+                        with self._completer_lock:
+                            self.completer.reset_completions()
+                        self.refresh_completions(persist_priorities='keywords')
                     elif query.meta_changed:
-                        self.refresh_completions(reset=False)
+                        self.refresh_completions(persist_priorities='all')
                     elif query.path_changed:
                         logger.debug('Refreshing search path')
                         with self._completer_lock:
@@ -346,15 +442,17 @@ class PGCli(object):
                         logger.debug('Search path: %r',
                                      self.completer.search_path)
 
+                # Allow PGCompleter to learn user's preferred keywords, etc.
+                with self._completer_lock:
+                    self.completer.extend_query_history(document.text)
+
                 self.query_history.append(query)
 
         except EOFError:
-            print ('Goodbye!')
-        finally:  # Reset the less opts back to original.
-            logger.debug('Restoring env var LESS to %r.', original_less_opts)
-            os.environ['LESS'] = original_less_opts
+            if not self.less_chatty:
+                print ('Goodbye!')
 
-    def _build_cli(self):
+    def _build_cli(self, history):
 
         def set_vi_mode(value):
             self.vi_mode = value
@@ -366,13 +464,17 @@ class PGCli(object):
         def prompt_tokens(_):
             return [(Token.Prompt, '%s> ' % self.pgexecute.dbname)]
 
+        def get_continuation_tokens(cli, width):
+            return [(Token.Continuation, '.' * (width - 1) + ' ')]
+
         get_toolbar_tokens = create_toolbar_tokens_func(
             lambda: self.vi_mode, self.completion_refresher.is_refreshing)
 
-        layout = create_default_layout(
-            lexer=PostgresLexer,
-            reserve_space_for_menu=True,
+        layout = create_prompt_layout(
+            lexer=PygmentsLexer(PostgresLexer),
+            reserve_space_for_menu=4,
             get_prompt_tokens=prompt_tokens,
+            get_continuation_tokens=get_continuation_tokens,
             get_bottom_toolbar_tokens=get_toolbar_tokens,
             display_completions_in_columns=self.wider_completion_menu,
             multiline=True,
@@ -383,15 +485,15 @@ class PGCli(object):
                    filter=HasFocus(DEFAULT_BUFFER) & ~IsDone()),
             ])
 
-        history_file = self.config['main']['history_file']
-        if history_file == 'default':
-            history_file = config_location() + 'history'
         with self._completer_lock:
             buf = PGBuffer(
                 always_multiline=self.multi_line,
                 completer=self.completer,
-                history=FileHistory(os.path.expanduser(history_file)),
-                complete_while_typing=Always())
+                history=history,
+                complete_while_typing=Always(),
+                accept_action=AcceptAction.RETURN_DOCUMENT)
+
+            editing_mode = EditingMode.VI if self.vi_mode else EditingMode.EMACS
 
             application = Application(
                 style=style_factory(self.syntax_style, self.cli_style),
@@ -399,14 +501,21 @@ class PGCli(object):
                 buffer=buf,
                 key_bindings_registry=key_binding_manager.registry,
                 on_exit=AbortAction.RAISE_EXCEPTION,
-                ignore_case=True)
+                on_abort=AbortAction.RETRY,
+                ignore_case=True,
+                editing_mode=editing_mode)
 
-            cli = CommandLineInterface(
-                application=application,
-                eventloop=create_eventloop())
+            cli = CommandLineInterface(application=application,
+                                       eventloop=self.eventloop)
 
             return cli
 
+    def _should_show_limit_prompt(self, status, cur):
+        """returns True if limit prompt should be shown, False otherwise."""
+        if not is_select(status):
+            return False
+        return self.row_limit > 0 and cur and cur.rowcount > self.row_limit
+
     def _evaluate_command(self, text):
         """Used to run a command entered by the user during CLI operation
         (Puts the E in REPL)
@@ -434,9 +543,8 @@ class PGCli(object):
             logger.debug("headers: %r", headers)
             logger.debug("rows: %r", cur)
             logger.debug("status: %r", status)
-            threshold = 1000
-            if (is_select(status) and
-                    cur and cur.rowcount > threshold):
+            threshold = self.row_limit
+            if self._should_show_limit_prompt(status, cur):
                 click.secho('The result set has more than %s rows.'
                             % threshold, fg='red')
                 if not click.confirm('Do you want to continue?'):
@@ -483,35 +591,64 @@ class PGCli(object):
             except OperationalError as e:
                 click.secho(str(e), err=True, fg='red')
 
-    def adjust_less_opts(self):
-        less_opts = os.environ.get('LESS', '')
-        self.logger.debug('Original value for LESS env var: %r', less_opts)
-        os.environ['LESS'] = '-SRXF'
+    def refresh_completions(self, history=None, persist_priorities='all'):
+        """ Refresh outdated completions
 
-        return less_opts
+        :param history: A prompt_toolkit.history.FileHistory object. Used to
+                        load keyword and identifier preferences
 
-    def refresh_completions(self, reset=False):
-        if reset:
-            with self._completer_lock:
-                self.completer.reset_completions()
+        :param persist_priorities: 'all' or 'keywords'
+        """
+
+        callback = functools.partial(self._on_completions_refreshed,
+                                     persist_priorities=persist_priorities)
+        c = self.config
+        settings = {'casing_file': get_casing_file(c),
+          'generate_casing_file': c['main'].as_bool('generate_casing_file'),
+          'asterisk_column_order': c['main']['asterisk_column_order']}
         self.completion_refresher.refresh(self.pgexecute, self.pgspecial,
-                                          self._on_completions_refreshed)
+            callback, history=history, settings=settings)
         return [(None, None, None,
                 'Auto-completion refresh started in the background.')]
 
-    def _on_completions_refreshed(self, new_completer):
-        self._swap_completer_objects(new_completer)
+    def _on_completions_refreshed(self, new_completer, persist_priorities):
+        self._swap_completer_objects(new_completer, persist_priorities)
 
         if self.cli:
             # After refreshing, redraw the CLI to clear the statusbar
             # "Refreshing completions..." indicator
             self.cli.request_redraw()
 
-    def _swap_completer_objects(self, new_completer):
+    def _swap_completer_objects(self, new_completer, persist_priorities):
         """Swap the completer object in cli with the newly created completer.
+
+            persist_priorities is a string specifying how the old completer's
+            learned prioritizer should be transferred to the new completer.
+
+              'none'     - The new prioritizer is left in a new/clean state
+
+              'all'      - The new prioritizer is updated to exactly reflect
+                           the old one
+
+              'keywords' - The new prioritizer is updated with old keyword
+                           priorities, but not any other.
         """
         with self._completer_lock:
+            old_completer = self.completer
             self.completer = new_completer
+
+            if persist_priorities == 'all':
+                # Just swap over the entire prioritizer
+                new_completer.prioritizer = old_completer.prioritizer
+            elif persist_priorities == 'keywords':
+                # Swap over the entire prioritizer, but clear name priorities,
+                # leaving learned keyword priorities alone
+                new_completer.prioritizer = old_completer.prioritizer
+                new_completer.prioritizer.clear_names()
+            elif persist_priorities == 'none':
+                # Leave the new prioritizer as is
+                pass
+
             # When pgcli is first launched we call refresh_completions before
             # instantiating the cli object. So it is necessary to check if cli
             # exists before trying the replace the completer object in cli.
@@ -541,10 +678,14 @@ class PGCli(object):
         help='database name to connect to.')
 @click.option('--pgclirc', default=config_location() + 'config',
         envvar='PGCLIRC', help='Location of pgclirc file.')
+@click.option('-D', '--dsn', default='', envvar='DSN',
+        help='Use DSN configured into the [alias_dsn] section of pgclirc file.')
+@click.option('-R', '--row-limit', default=None, envvar='PGROWLIMIT', type=click.INT,
+        help='Set threshold for row limit prompt. Use 0 to disable prompt.')
 @click.argument('database', default=lambda: None, envvar='PGDATABASE', nargs=1)
 @click.argument('username', default=lambda: None, envvar='PGUSER', nargs=1)
 def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
-        username, version, pgclirc):
+        username, version, pgclirc, dsn, row_limit):
 
     if version:
         print('Version:', __version__)
@@ -566,13 +707,24 @@ def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
             print ('Please move the existing config file ~/.pgclirc to',
                    config_full_path)
 
-    pgcli = PGCli(prompt_passwd, never_prompt, pgclirc_file=pgclirc)
+    pgcli = PGCli(prompt_passwd, never_prompt, pgclirc_file=pgclirc,
+                  row_limit=row_limit)
 
     # Choose which ever one has a valid value.
     database = database or dbname
     user = username or user
 
-    if '://' in database:
+    if dsn is not '':
+        try:
+            cfg = load_config(config_full_path)
+            dsn_config = cfg['alias_dsn'][dsn]
+        except:
+            click.secho('Invalid DSNs found in the config file. '\
+                'Please check the "[alias_dsn]" section in pgclirc.',
+                 err=True, fg='red')
+            exit(1)
+        pgcli.connect_uri(dsn_config)
+    elif '://' in database:
         pgcli.connect_uri(database)
     elif "=" in database:
         pgcli.connect_dsn(database)
@@ -613,7 +765,7 @@ def format_output(title, cur, headers, status, table_format, expanded=False, max
         else:
             tabulated, rows = tabulate(cur, headers, tablefmt=table_format,
                 missingval='<null>')
-            if (max_width and
+            if (max_width and rows and
                     content_exceeds_width(rows[0], max_width) and
                     headers):
                 output.append(expanded_table(rows, headers))
