@@ -3,7 +3,7 @@ import re
 import sqlparse
 from collections import namedtuple
 from sqlparse.sql import IdentifierList, Identifier, Function
-from sqlparse.tokens import Keyword, DML, Punctuation, Token
+from sqlparse.tokens import Keyword, DML, Punctuation, Token, Error
 
 cleanup_regex = {
         # This matches only alphanumerics and underscores.
@@ -46,6 +46,8 @@ def last_word(text, include='alphanum_underscore'):
     '\\\\def;'
     >>> last_word('bac::def', include='most_punctuations')
     'def'
+    >>> last_word('"foo*bar', include='most_punctuations')
+    '"foo*bar'
     """
 
     if not text:   # Empty string
@@ -64,6 +66,9 @@ def last_word(text, include='alphanum_underscore'):
 
 TableReference = namedtuple('TableReference', ['schema', 'name', 'alias',
                                                'is_function'])
+TableReference.ref = property(lambda self: self.alias or (
+  self.name if self.name.islower() or self.name[0] == '"'
+  else '"' + self.name + '"'))
 
 
 # This code is borrowed from sqlparse example script.
@@ -101,12 +106,14 @@ def extract_from_part(parsed, stop_at_punctuation=True):
             elif item.ttype is Keyword and (
                     not item.value.upper() == 'FROM') and (
                     not item.value.upper().endswith('JOIN')):
-                raise StopIteration
+                tbl_prefix_seen = False
             else:
                 yield item
-        elif ((item.ttype is Keyword or item.ttype is Keyword.DML) and
-                item.value.upper() in ('COPY', 'FROM', 'INTO', 'UPDATE', 'TABLE', 'JOIN',)):
-            tbl_prefix_seen = True
+        elif item.ttype is Keyword or item.ttype is Keyword.DML:
+            item_val = item.value.upper()
+            if (item_val in ('COPY', 'FROM', 'INTO', 'UPDATE', 'TABLE') or
+                    item_val.endswith('JOIN')):
+                tbl_prefix_seen = True
         # 'SELECT a, FROM abc' will detect FROM as part of the column list.
         # So this check here is necessary.
         elif isinstance(item, IdentifierList):
@@ -120,6 +127,31 @@ def extract_from_part(parsed, stop_at_punctuation=True):
 def extract_table_identifiers(token_stream, allow_functions=True):
     """yields tuples of TableReference namedtuples"""
 
+    # We need to do some massaging of the names because postgres is case-
+    # insensitive and '"Foo"' is not the same table as 'Foo' (while 'foo' is)
+    def parse_identifier(item):
+        name = item.get_real_name()
+        schema_name = item.get_parent_name()
+        alias = item.get_alias()
+        if not name:
+            schema_name = None
+            name = item.get_name()
+            alias = alias or name
+        schema_quoted = schema_name and item.value[0] == '"'
+        if schema_name and not schema_quoted:
+            schema_name = schema_name.lower()
+        quote_count = item.value.count('"')
+        name_quoted = quote_count > 2 or (quote_count and not schema_quoted)
+        alias_quoted = alias and item.value[-1] == '"'
+        if alias_quoted or name_quoted and not alias and name.islower():
+            alias = '"' + (alias or name) + '"'
+        if name and not name_quoted and not name.islower():
+            if not alias:
+                alias = name
+            name = name.lower()
+        return schema_name, name, alias
+
+
     for item in token_stream:
         if isinstance(item, IdentifierList):
             for identifier in item.get_identifiers():
@@ -136,20 +168,13 @@ def extract_table_identifiers(token_stream, allow_functions=True):
                     yield TableReference(schema_name, real_name,
                                          identifier.get_alias(), is_function)
         elif isinstance(item, Identifier):
-            real_name = item.get_real_name()
-            schema_name = item.get_parent_name()
+            schema_name, real_name, alias = parse_identifier(item)
             is_function = allow_functions and _identifier_is_function(item)
 
-            if real_name:
-                yield TableReference(schema_name, real_name, item.get_alias(),
-                                     is_function)
-            else:
-                name = item.get_name()
-                yield TableReference(None, name, item.get_alias() or name,
-                                     is_function)
+            yield TableReference(schema_name, real_name, alias, is_function)
         elif isinstance(item, Function):
-            yield TableReference(None, item.get_real_name(), item.get_alias(),
-                                 allow_functions)
+            schema_name, real_name, alias = parse_identifier(item)
+            yield TableReference(None, real_name, alias, allow_functions)
 
 
 # extract_tables is inspired from examples in the sqlparse lib.
@@ -161,7 +186,7 @@ def extract_tables(sql):
     """
     parsed = sqlparse.parse(sql)
     if not parsed:
-        return []
+        return ()
 
     # INSERT statements must stop looking for tables at the sign of first
     # Punctuation. eg: INSERT INTO abc (col1, col2) VALUES (1, 2)
@@ -177,7 +202,9 @@ def extract_tables(sql):
     # to have is_function=True
     identifiers = extract_table_identifiers(stream,
                                             allow_functions=not insert_stmt)
-    return list(identifiers)
+    # In the case 'sche.<cursor>', we get an empty TableReference; remove that
+    return tuple(i for i in identifiers if i.name)
+
 
 
 def find_prev_keyword(sql):
@@ -254,6 +281,29 @@ def _parsed_is_open_quote(parsed):
     return False
 
 
+def parse_partial_identifier(word):
+    """Attempt to parse a (partially typed) word as an identifier
+
+    word may include a schema qualification, like `schema_name.partial_name`
+    or `schema_name.` There may also be unclosed quotation marks, like
+    `"schema`, or `schema."partial_name`
+
+    :param word: string representing a (partially complete) identifier
+    :return: sqlparse.sql.Identifier, or None
+    """
+
+    p = sqlparse.parse(word)[0]
+    n_tok = len(p.tokens)
+    if n_tok == 1 and isinstance(p.tokens[0], Identifier):
+        return p.tokens[0]
+    elif p.token_next_match(0, Error, '"'):
+        # An unmatched double quote, e.g. '"foo', 'foo."', or 'foo."bar'
+        # Close the double quote, then reparse
+        return parse_partial_identifier(word + '"')
+    else:
+        return None
+
+
 if __name__ == '__main__':
     sql = 'select * from (select t. from tabl t'
     print (extract_tables(sql))
