@@ -1,4 +1,5 @@
 from pygments.token import Token
+from prompt_toolkit.enums import DEFAULT_BUFFER
 
 
 def create_toolbar_tokens_func(get_vi_mode_enabled, get_is_refreshing):
@@ -13,17 +14,17 @@ def create_toolbar_tokens_func(get_vi_mode_enabled, get_is_refreshing):
         result = []
         result.append((token, ' '))
 
-        if cli.buffers['default'].completer.smart_completion:
+        if cli.buffers[DEFAULT_BUFFER].completer.smart_completion:
             result.append((token.On, '[F2] Smart Completion: ON  '))
         else:
             result.append((token.Off, '[F2] Smart Completion: OFF  '))
 
-        if cli.buffers['default'].always_multiline:
+        if cli.buffers[DEFAULT_BUFFER].always_multiline:
             result.append((token.On, '[F3] Multiline: ON  '))
         else:
             result.append((token.Off, '[F3] Multiline: OFF  '))
 
-        if cli.buffers['default'].always_multiline:
+        if cli.buffers[DEFAULT_BUFFER].always_multiline:
             result.append((token,
                 ' (Semi-colon [;] will end the line)'))
 
